/*
 * _857702.c
 *
 * Created: 14/09/2015 19:13:39
 *  Author: German
 *
 *  Luces de matrcula y cabinas de la 7702:
 *      F5 enciende cabina delantera
 *      F6 enciende la cabina trasera
 *      F7 enciende las matrculas
 *      Con la locomotora en marcha, la cabina delantera si est encendida, reduce su iluminacin al mnimo
*/
// cdigo similar al de '85_700_cabina'

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>

#define F_CPU 8000000UL

// pines empleados:
#define cab0    PB0     // cabina de cabeza
#define cab1    PB1     // cabina de cola
#define matr    PB4     // matrculas
#define p_dcc   PB3

#define BR_MIN  15      // brillo mnimo
#define BR_MAX  100     // brillo mximo

// direcciones de las cvs usadas
#define CV_1    0       // [P]+[*]
#define CV_17  16
#define CV_18  17
#define CV_29  28       // [P]+[STOP] CV29 debera coger la de la mquina
// hay que grabar la direccin 0 con el valor 3, y la 28 con 0 (por ejemplo) de la EEPROM
//   o bien con los valores que tenga en ellas la locomotora donde se instalar
// una vez instalado grabar estas cvs con lo que se programe para la locomotora

// Funciones DCC:
//  '1 0 0 F0  F4  F3  F2  F1'  Function Group One Instruction (100)
#define F0     16       // pesos de cada funcin en el byte de control
#define F1      1
#define F2      2
#define F3      4
#define F4      8
//  '1 0 1  1  F8  F7  F6  F5'  Function Group Two Instruction (101)
//  '1 0 1  0  F12 F11 F10 F9'  Function Group Two Instruction (101)
#define F5      1
#define F6      2
#define F7      4
#define F8      8
#define F9     16
#define F10    32
#define F11    64
#define F12   128

#define SI 0            // para el prembulo, la primera vez de la interrupcin
#define NO 1

typedef unsigned char byte;

// control de las funciones, una en cada bit:
//      ctrl_fn1 ->  0   0   0  F0    F4 F3 F2 F1
//      ctrl_fn2 -> F12 F11 F10 F9    F8 F7 F6 F5
volatile byte ctrl_fn1, ctrl_fn2, sentido, vel_loco;
byte buf_dcc[6];                        // buffer del ltimo comando recibido
byte cv_1, cv_17, cv_18, cv_29;         // cv_29 usa bit0=sentido y bit5=direccin larga/corta
volatile byte flag_16ms;                // flag activado por la int. del timer0 para la temporizacin

// interrupcin externa por cambio de nivel para decodififcar la seal DCC
// Cuenta el tiempo del nivel alto con el timer1
ISR(PCINT0_vect)
{
register byte bit_dcc, tmp_pulso;
static byte preamb;
static byte aux_dcc, x_or, idx, num_bits, modo_svc;
    if (PINB & (1<<p_dcc)) TCCR1 |= 4<<CS10;        // cuando se pone a nivel alto, arranca cuenta prescaler 1/8
    else                                            // cuando pasa a nivel bajo
    {
        TCCR1 &= ~(4<<CS10);                        // detiene la cuenta
        tmp_pulso = TCNT1;                          // lee los microsegundos (byte bajo)
        TCNT1 = 0;                                  // pone a cero el timer
        if (tmp_pulso>90) bit_dcc = 0;              // sin buscar exactitud: duracin mayor de 90 us => cero
        else bit_dcc = 1;                           //    y menor uno
        if (preamb == SI)                           // prembulo
        {
            if (bit_dcc)                            // otro '1'
            {if (num_bits) num_bits--;}
            else
            {
                if (!num_bits)                      // el '0' de inicio de datos
                {preamb = NO; num_bits = 9; x_or = 0; idx = 0;}
                else                                // no se cumple la trama, vuelta a inicio
                num_bits = 10;
            }
        }
        else                // recepcin de los bytes
        {
            if (--num_bits) aux_dcc = aux_dcc * 2 + bit_dcc;
            else                                    // el separador de bytes
            {
                if (!bit_dcc)                       // cero, a por el siguiente byte
                {
                    buf_dcc [idx++] = aux_dcc;
                    x_or ^= aux_dcc;                // actualiza la x_or
                    num_bits = 9;
                }
                else                                // uno, fin del paquete
                {
                    preamb = SI;
                    num_bits = 10;                  // vuelta a empezar
                    if (x_or == aux_dcc)             // paquete correcto
                    {
                        // Particin de direcciones (RP-921):
                        // Address 00000000 (0): Broadcast address
                        // Addresses 00000001-01111111 (1-127)(inclusive): Multi-Function decoders with 7 bit addresses
                        // Addresses 10000000-10111111 (128-191)(inclusive): Basic Accessory Decoders with 9 bit addresses
                        //     and Extended Accessory Decoders with 11-bit addresses
                        // Addresses 11000000-11100111 (192-231)(inclusive): Multi Function Decoders with 14 bit addresses
                        // Addresses 11101000-11111110 (232-254)(inclusive): Reserved for Future Use
                        // Address 11111111 (255): Idle Packet
                        if (!buf_dcc[0]) modo_svc = 1;                                  // direccin 0 -> reset
                        else
                        {
                            if ((buf_dcc[0] == cv_1) && !(cv_29 & 0x20))                // direccin corta
                            {
                                if (buf_dcc[1] == 0x3f)                                 // 128 pasos -> SVVV VVVV
                                {
                                    vel_loco = ((buf_dcc[2] & 0x7f) * 2)/9;             // convierte 128 -> 28 pasos aprox.
                                    if (buf_dcc[2] & 0x80) sentido = -1;
                                    else sentido = 0;
                                    if (cv_29 & 1) sentido = ~sentido;
                                }
                                else                                                    // 28 pasos
                                {
                                    if ((buf_dcc[1] & 0xc0) == 0x40)                    // formato: 01sD DDDD
                                    {
                                        vel_loco = (buf_dcc[1] & 0x0f)*2;               // 28 pasos de velocidad
                                        if ((buf_dcc[1] & 0x10) == 0x10) vel_loco++;    // se aade el bit de menor peso
                                        if (vel_loco>3) vel_loco -= 3;                  // para dejar los pasos de 1 a 28
                                        else vel_loco = 0;                              //  y si no, 0 para las luces
                                        if (buf_dcc[1] & 0x20) sentido = -1;
                                        else sentido = 0;
                                        if (cv_29 & 1) sentido = ~sentido;
                                    }
                                }
                                if ((buf_dcc[1] & 0xe0) == 0x80)                        // Function Group One:  100F0 F4F3F2F1
                                ctrl_fn1 = buf_dcc[1] & 0x1f;                           // funciones
                                else if ((buf_dcc[1] & 0xf0) == 0xa0)                   // funciones grupo 2: 101S DDDD (S=1->D8..5 // S=0->D12..9)
                                ctrl_fn2 = (ctrl_fn2 & 0x0f) | ((buf_dcc[1] & 0x0f) << 4); // D12..D9 en nibble alto
                                else if ((buf_dcc[1] & 0xf0) == 0xb0)                   // D8..D5
                                ctrl_fn2 = (ctrl_fn2 & 0xf0) | (buf_dcc[1] & 0x0f);     // pone D8..D5 en nibble bajo
                            }
                            // direccin de decodificador de accesorios o prog modo directo
                            /*                            else if ((buf_dcc[0] == 0x74) && modo_svc)      // verificacin
                            {
                            if (((buf_dcc[1]==CV_1) && (buf_dcc[2]==cv_1))  \
                            || ((buf_dcc[1]==CV_17) && (buf_dcc[2]==cv_17)) \
                            || ((buf_dcc[1]==CV_18) && (buf_dcc[2]==cv_18)) \
                            || ((buf_dcc[1]==CV_29) && (buf_dcc[2]==cv_29)))
                            {
                            PORTB = 1; while(1);                               // intento de ack ***** aqu no se pondr (lo har el de la loco)
                            }
                            }*/
                            else if ((buf_dcc[0] == 0x7c) && modo_svc)      // escritura
                            {   // graba si es alguna de las cvs que utiliza
                                if ((buf_dcc[1]==CV_1) || (buf_dcc[1]==CV_17) || (buf_dcc[1]==CV_18) || (buf_dcc[1]==CV_29))
                                {
                                    eeprom_update_byte((uint8_t *)((uint16_t)buf_dcc[1]), buf_dcc[2]);
                                    PORTB = 1; while(1);
                                }
                            }
                            else if ((buf_dcc[0] != 0xff) && (cv_29 & 0x20))            // entre 192 y 254: direccin larga
                            {
                                if ((buf_dcc[0] == cv_17) && (buf_dcc[1] == cv_18))     // es su direccin
                                {
                                    if (buf_dcc[2] == 0x3f)                             // 128 pasos -> SVVV VVVV
                                    {
                                        vel_loco = ((buf_dcc[3] & 0x7f) * 2)/9;         // convierte 128->28 aprox.
                                        if (buf_dcc[3] & 0x80) sentido = -1;            // sentido
                                        else sentido = 0;
                                        if (cv_29 & 1) sentido = ~sentido;
                                    }
                                    else                                                // 28 pasos
                                    {
                                        if ((buf_dcc[2] & 0xc0) == 0x40)                // formato: 01sD DDDD
                                        {
                                            vel_loco = (buf_dcc[2] & 0x0f)*2;           // 28 pasos de velocidad
                                            if ((buf_dcc[2] & 0x10) == 0x10) vel_loco++;   // se aade el bit de menor peso
                                            if (vel_loco>3) vel_loco -= 3;              // para dejar los pasos de 1 a 28
                                            else vel_loco = 0;                          //  y si no, 0 para las luces
                                            if (buf_dcc[2] & 0x20) sentido = -1;
                                            else sentido = 0;
                                            if (cv_29 & 1) sentido = ~sentido;
                                        }
                                    }
                                    if ((buf_dcc[2] & 0xe0) == 0x80)                    // Function Group One:  100F0 F4F3F2F1
                                    ctrl_fn1 = buf_dcc[2] & 0x1f;                       // funciones
                                    else if ((buf_dcc[2] & 0xf0) == 0xa0)               // funciones grupo 2: 101S DDDD (S=1->D8..5 // S=0->D12..9)
                                    ctrl_fn2 = (ctrl_fn2 & 0x0f) | ((buf_dcc[2] & 0x0f) << 4); // D12..D9 en nibble alto
                                    else if ((buf_dcc[2] & 0xf0) == 0xb0)                       // D8..D5
                                    ctrl_fn2 = (ctrl_fn2 & 0xf0) | (buf_dcc[2] & 0x0f); // pone D8..D5 en nibble bajo
                                }
                            }
                            modo_svc = 0;                                               // modo servicio desactivado
                        }
                    }
                }
            }
        }
    }
}

// prescaler 1/256, luego cada 16320 us, por ser phase correct (cuenta up y down)
ISR(TIM0_OVF_vect, ISR_NOBLOCK)
{
    flag_16ms = 1;
}

// configuracin inicial para ATtiny85
void setup()
{
    DDRB = (1<<cab0) | (1<<cab1) | (1<<matr);   // PB0 y PB1 salida
    PORTB = 0b00100;                // pines no utilizados son entradas con pull-up
    GIMSK |= (1<<PCIE);             // habilita interrupcin por cambio de nivel
    PCMSK |= (1<<PCINT3);           // interrupcin por cambio de PB3
    // mode #1 pwm, Phase correct:
    TCCR0A = (2<<COM0A0) | (2<<COM0B0) | (1<<WGM00);
    OCR0A = 0;                      // inicialmente parado
    OCR0B = 0;
    TCCR0B |= (4 << CS00);          // prescaler 1/256
    TIMSK |= (1 << TOIE0);          // habilita interrupcin por overflow (temporizacin)

    cv_1  = eeprom_read_byte((uint8_t*) CV_1);
    cv_17 = eeprom_read_byte((uint8_t*) CV_17);
    cv_18 = eeprom_read_byte((uint8_t*) CV_18);
    cv_29 = eeprom_read_byte((uint8_t*) CV_29);

    sei();                          // habilitador general
}


int main(void)
{
    setup();
    while(1)
    {
        // opera las funciones usadas
        if (flag_16ms)                      // cada 16 ms
        {
            flag_16ms = 0;
            if (ctrl_fn2 & F7)              // luces de matrcula
                PORTB |= (1 << matr);       // encendidas
            else
                PORTB &= ~(1 << matr);      // apagadas
            // luces cabinas:
            if (sentido)                    // avance
            {
                if (ctrl_fn2 & F5)          // cabina delantera encendida
                {
                    if (vel_loco)           // locomotora en marcha
                    {
                        if (OCR0A > BR_MIN) OCR0A-=2;       // disminuye brillo hasta mnimo
                        else if (!OCR0A) OCR0A = BR_MIN;    // si estaba apagada la enciende
                    }
                    else
                        OCR0A = BR_MAX;     // parada luce al mximo
                }
                else
                    OCR0A = 0;              // apagada

                if (ctrl_fn2 & F6)          // cabina trasera encendida
                    OCR0B = BR_MAX;
                else
                    OCR0B = 0;              // cabina trasera apagada
            }
            else                            // retroceso
            {
                if (ctrl_fn2 & F5)          // cabina delantera encendida
                {
                    if (vel_loco)           // locomotora en marcha
                    {
                        if (OCR0B > BR_MIN) OCR0B-=2;       // disminuye brillo hasta mnimo
                        else if (!OCR0B) OCR0B = BR_MIN;    // si estaba apagada la enciende
                    }
                    else
                    OCR0B = BR_MAX;     // parada luce al mximo
                }
                else
                    OCR0B = 0;              // apagada

                if (ctrl_fn2 & F6)          // cabina trasera encendida
                    OCR0A = BR_MAX;
                else
                    OCR0A = 0;              // cabina trasera apagada
            }
        }
    }
}